##############################################################################################################
#									
# Routine :	alm_copyproject.pl	v1.27			
#   									
# Description :								
#	This perl script copies ALM infrastructure records and their links for a given ALM project.This script clones 
#	a Project and its Phases, Roles, WorkConfigurations, DefaultRequest,and DefaultTask.		
#									
# Syntax :								
#	ratlperl alm_copyproject.pl {-projadmin <projadmin>} {-pw <pw>} {-dbname <db>} {-dbset <dbset>} {-projectid <projectid>}
#       [-clonedprojname [<clonedprojname>]] [-category [<category>]] [-release [<release>]]
#	example: ratlperl alm_copyproject.pl -projadmin ProjectAdmin -pw "" -dbname ALM -dbset CQMS.ALMM3.GALWAY -project ALM00001015
#	-clonedprojname CopyTest -category TestCat -release TestRel	
#									
# Output :								
#	New project is copied from the old project.					
#														
#									
# Licensed Materials - Property of IBM
# Rational ClearQuest
# (C) Copyright IBM Corp. 2008-2009.  All Rights Reserved
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp
# 											
##############################################################################################################
#
# Change Log:
#
# Ver       CRs                 Date         Programmer        Comments (fixed, fixed with dependency, added feature)
# 1.27      RATLC01271668	04/28/2008   majidi	       Better error handling	
# 1.26	    Misc	        12/10/2008   majidi	       Added support for Submitter, SubmitDate,DueDate,Keywords,StatusLabel,Originator	
# 1.25      Misc		11/18/2008   majidi	       ALM 1.2 fixes (Error reporting improvement, PrimaryRole support in WC)
# 1.24      RATLC01234428	08/01/2008   majidi	       CopyProject should not copy additional Tasks associated with DefaultRequest
#           RATLC01229064       08/01/2008   majidi	       Restructure the copy script to make maintenance easy
# 
# There is change log gap here. Some of the fixes (handful) went directly to the schema record script.
#
# 1.23      RATLC01225506	03/17/2008   majidi	       Set the PriorPhaseLabel to set the PriorPhase (change in m3it4)
# 1.22      offline rfe		03/05/2008   majidi	       Added time stamp to project name with no clonedprojectname
# 1.21      offline defect	03/03/2008   majidi	       Groups field copy support in Role added (post M3 change)
# 1.20 	    offline defect	02/29/2008   majidi	       ApproveActions are now copied from the source role
# 							       instead of, from the rolelabel ( by value changed hook)
# 1.19      offline task	02/24/2008   majidi	       Get replica from session plus various minor changes
# 1.18      offline rfe 	02/19/2008   majidi	       Changed project prefix from Cloned_ to CopyOf_ 
# 1.17      offline rfe 	02/13/2008   majidi	       Set SecurityPolicy for cloned records to
#							       Project->SecurityPolicy except (Role->Everyone)
# 1.16      RATLC01049138 	02/07/2008   majidi	       Project->Name, Category and Release be input parameters 
# 1.15      RATLC00398015	01/30/2008   majidi	       Iterations records are not cloned
#							       Also see RATLC00398015 (dependency)
# 1.14      offline defect	01/31/2008   majidi            Prior phase/iteration does not work
# 1.13      RATLC01049975       01/29/2008   majidi	       Error when Project has CurrentPhaseLabel set
# 1.12	    offline defect	01/28/2008   majidi	       Error when DefaultTask is empty
# 1.11	    RATLC01049252     	01/28/2008   majidi	       Clone DefaultRequest and DefaultTask on the project	
# 	    RATLC01049178     	01/28/2008   majidi	       Role members are missing (CopyEntity defect)	
# 1.1	    RATLC01220362	01/22/2008   majidi	       WorkConfiguration record does not clone in ALMM3
# 1.0	    RATLC01220362	01/20/2008   majidi	       Added support for ALM schema infrastructure cloning
# 	    	                             Shireen           Created the CopyEntity demo script
#
# Known issues:
# 
# 
##############################################################################################################
#!ratlperl -l -w
use CQPerlExt;
use strict;
use File::Basename;
use Getopt::Long;

######################
my $progname = basename $0;
my $entity;
my $VERSION="1.27";
my $CopyToProjectID;
# Define the input parameters
my @options = qw(
    projadmin=s		
    pw=s		
    dbname=s		
    dbset=s		
    projectid=s
    clonedprojname:s
    category:s
    release:s
    secpolicy:s	
    help);

my %options = (
    projadmin => undef,
    pw => undef,
    dbname => undef,
    dbset => undef,
    projectid => undef,
    clonedprojname => undef,
    category => undef,
    release => undef,
    secpolicy => undef,
);

######################
# Get the input parameters
unless (GetOptions(\%options, @options)) {
 	print "ʾ÷";	
      	my $retval = &Usage;
      	exit 0;

}
######################
# Display the help usage
if ($options{help}) {
	print "-projadmin       ProjectAdmin ĵ¼ʶ\n";
	print "-pw              ProjectAdmin 롣\n";
	print "-dbname          ûݿơ\n";
	print "-dbset           ݿݿ⼯ӵơ\n";
	print "-projectid       Ҫ¡Ŀıʶ\n";
	print "-clonedprojname  ¡Ŀơ\n";
	print "-category        ¡Ŀ ALMCategory\n";
	print "-release         ¡Ŀ ALMRelease\n";
	print "-help            ڰĿѡ־\n";
	exit 0;
}

######################
# End program if any of the parameters are not defined
  if (!defined $options{projadmin}
     or !defined $options{pw}
     or !defined $options{dbname}
     or !defined $options{dbset}
     or !defined $options{projectid}) 
  { 
     #print"ʾ÷";
     my $retval = &Usage();
     exit 1;
  }

######################

my $session = CQSession::Build();
eval{$session->UserLogon($options{projadmin},$options{pw},$options{dbname},$options{dbset});};
die "¼ʱ $options{projadmin},$options{pw},$options{dbname},$options{dbset} \n$@\n" if ($@);


# Need a Entity Object for project
eval {$entity = $session->GetEntity("ALMProject","$options{projectid}");};
if ($@){

	CQPerlExt::CQSession_Unbuild($session);
	die " GetEntity $options{projectid} յ\n$@\n";
}

my $currUser = $session->GetUserLoginName();
almproject_ALM_CopyProject("$options{clonedprojname}\n$options{category}\n$options{release}");
$CopyToProjectID=$session->GetNameValue("CopyToProjectID");
$session->SetNameValue("CopyToProjectID", "");
print("ƳɹµĿʶΪ$CopyToProjectID\n");
CQPerlExt::CQSession_Unbuild($session);
exit 0;
######################
# Subroutine to display the usage
sub Usage {
	 print "\n$progname $VERSIONڸ ALM Ŀʵó\n
	
         $progname
	 {-projadmin <ĿԱ>}
         {-pw <>}
         {-dbname <ݿ>}
         {-dbset <ݿ⼯>}
         {-projectid <Ŀʶ>}
	 [-clonedprojname [<¡Ŀ>]]
 	 [-category [<>]]
 	 [-release [<а>]]
	 [-help]
";
return 0;
}

#####################################################################################
# The following will be copied as it is to the record script in the schema. Make sure
# these two subs are identical.
#####################################################################################
sub almproject_ALM_CopyProject {
    my($result);
    my($param) = @_;
    # record type name is ALMProject
	
	# This function implements ClearQuest's CopyEntity API
	# Please note that the aforementioned API is called internally only.
	# It is neither published nor supported, and this implementation should not be imitated.

	### Define variables
	my @params = split(/\n/, $param);

	my $projEntity;
	my @copyArray=();
	my @skipArray=();
	my $idx;
	my $dateStr;
	my $entsCount;
	my $status;
	my $strarray1;
	my $strarray2;
	my $ent;
	my $dspname;
	my $clonedProjName;
	my %AWCHash=();
	my $refh_AWC;
	my $primary;
	my $secondary;
	my $roles;
	my $primaryRole;
	my @ptypes=();
	my @stypes=();
	my @AWCroles=();
	my $AWCEntity;
	my $clonedProjID;
	my $currentPhaseLabel;
	my $phaseLabel;
	my $finaldspname;
	my $projName;
	my $localreplica;
	my $adminSess; 
	my $clonedRequestID;
	my $clonedTaskID;
	my $projEntity;
	my $taskOwner;
	my $taskType;
	my $taskEntity;
	my $taskRoles;
	my %ARHash=();
	my $refh_AR;
	my $members;
	my $AREntity;
	my @ARMembers=();
	my $oldmembers=();
	my $origdspName;
	my $cloneddspName;
	my %AIHash=();
	my $refh_AI;
	my $AIEntity;
	my $phase;
	my $iterationLabel;
	my $priorPhaseLabel;
	my %APHash=();
	my $refh_AP;
	my $APEntity;
	my $priorIteration;
	my $projectID;
	my $clonedprojSecPol;
	my %opts=();
	my $failure;
	my $entityobjs;
	my $actions;
	my @ARActions=();
	my @ARGroups=();
	my $groups;
	my ($querydef,$operator,$resultset,$count);
	my $hasParam = 0;

	use subs qw(CQValidateCommit);
	use subs qw(GetCurrentDate);

	### Initialize some variables
	$projEntity=$entity;
	$session->SetNameValue("CopyToProjectID", "");
	$localreplica= ($session->GetLocalReplica())->GetDisplayName(); 
	$projectID = $projEntity->GetFieldValue("id")->GetValue();
	
	if ($#params > 0) # Get the parameters
	{
		$hasParam = 1;
		$opts{clonedprojname}=$params[0];
		$opts{category}= $params[1];
		$opts{release}= $params[2];

	}
	else			# Get the paramaters from the Project form
	{
		$opts{clonedprojname}=$projEntity->GetFieldValue("CopyToProjectName")->GetValue();
		$opts{category}= $projEntity->GetFieldValue("CopyToCategory")->GetValue();
		$opts{release}= $projEntity->GetFieldValue("CopyToRelease")->GetValue();
	}

	# Check if there already exists an ALMProject with the same Name
	if (! ($opts{clonedprojname} eq "" or $opts{clonedprojname} =~ /^\s+$/))
	{
		$querydef = $session->BuildQuery("ALMProject");	
		$querydef->BuildField("dbid");
		$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
		$operator->BuildFilter("Name", $CQPerlExt::CQ_COMP_OP_EQ, [$opts{clonedprojname}]);
		$resultset = $session->BuildResultSet($querydef);
		$resultset->EnableRecordCount();
		$resultset->Execute;
		$count = $resultset->GetRecordCount();	
		if ($count > 0) { 
			CQPerlExt::CQSession_Unbuild($session);
			die "*** Ѵһ ALMProject ¼\n\t ƣ$opts{clonedprojname}\n"; 
		}		
	}

	# Check if there already exists an ALMProject with the same Category + Release combination
		$querydef = $session->BuildQuery("ALMProject");	
		$querydef->BuildField("dbid");
		$operator = $querydef->BuildFilterOperator($CQPerlExt::CQ_BOOL_OP_AND);
		
		if ($opts{category} eq "" or $opts{category} =~ /^\s+$/) { 
			CQPerlExt::CQSession_Unbuild($session);
			die "*** Ҫ ALMProjectҪһ***\n"; 
		}
		else {$operator->BuildFilter("Category", $CQPerlExt::CQ_COMP_OP_EQ, [$opts{category}]);}
		

		if ($opts{release} eq "" or $opts{release} =~ /^\s+$/)
		{$operator->BuildFilter("Release", $CQPerlExt::CQ_COMP_OP_IS_NULL, []);}
		else {$operator->BuildFilter("Release", $CQPerlExt::CQ_COMP_OP_EQ, [$opts{release}]);}

		$resultset = $session->BuildResultSet($querydef);
		$resultset->EnableRecordCount();
		$resultset->Execute;
		$count = $resultset->GetRecordCount();	
		if ($count > 0)	{ 
			CQPerlExt::CQSession_Unbuild($session);
			die "*** Ѵһ ALMProject ¼\n\t $opts{category}\n\t а棺$opts{release}\n"; 
		}


	######################
	$session->OutputDebugString("ڵ CopyEntity $@\n");

	##Copy Array
	##Temporary stopping iteration record cloning

	@copyArray=('ALMProject'
	,'ALMProject.Phases'
	,'ALMProject.Phases.Project'
	,'ALMProject.Roles'
	,'ALMProject.Roles.Project'
	,'ALMProject.WorkConfigurations'
	,'ALMProject.WorkConfigurations.ALMProject'
	,'ALMProject.DefaultRequest'
	,'ALMProject.DefaultRequest.Project'
	,'ALMProject.DefaultTask'
	,'ALMProject.DefaultTask.Project');

	##Skip Array

	@skipArray = ('ALMProject.SubProjects'
	,'ALMProject.SuperProjects'
	,'ALMProject.Release'
	,'ALMProject.CurrentPhaseLabel'
	,'ALMProject.PriorProject'
	,'ALMProject.NextProject'
	,'ALMProject.CurrentPhase'
	,'ALMProject.UCM_Project'
	,'ALMProject.StatusLabel'
	,'ALMProject.ReportTagsLabel'
	,'ALMProject.Phases.NextPhase'
	,'ALMProject.Phases.CurrentIteration'
	,'ALMProject.Phases.CurrentIterationLabel'
	,'ALMProject.Phases.StatusLabel'
	,'ALMProject.DefaultRequest.NotifyList'
	,'ALMProject.DefaultRequest.RequestsRelated'
	,'ALMProject.DefaultRequest.RequestsRelatedBack'
	,'ALMProject.DefaultRequest.CategoryPath'
	,'ALMProject.DefaultRequest.Phase'
	,'ALMProject.DefaultRequest.PhaseLabel'
	,'ALMProject.DefaultRequest.IterationLabel'
	,'ALMProject.DefaultRequest.Iteration'
	,'ALMProject.DefaultRequest.Comments'
	,'ALMProject.DefaultRequest.Attachement'
	,'ALMProject.DefaultRequest.BTBuild'
	,'ALMProject.DefaultRequest.CategroyTypeLabel'
	,'ALMProject.DefaultRequest.Tasks'
	,'ALMProject.DefaultRequest.StatusLabel'
	,'ALMProject.DefaultRequest.Keywords'
	,'ALMProject.DefaultRequest.Submitter'
	,'ALMProject.DefaultRequest.Originator'	
	,'ALMProject.DefaultTask.PhaseLabel'
	,'ALMProject.DefaultTask.NotifyList'
	,'ALMProject.DefaultTask.TasksRelated'
	,'ALMProject.DefaultTask.TasksRelatedBack'
	,'ALMProject.DefaultTask.Iteration'
	,'ALMProject.DefaultTask.Phase'
	,'ALMProject.DefaultTask.IterationLabel'
	,'ALMProject.DefaultTask.Activities'
	,'ALMProject.DefaultTask.Roles'
	,'ALMProject.DefaultTask.Comments'
	,'ALMProject.DefaultTask.Attachement'
	,'ALMProject.DefaultTask.BTBuild'
	,'ALMProject.DefaultTask.ResolutionCode'
	,'ALMProject.DefaultTask.Request'
	,'ALMProject.DefaultTask.StatusLabel'
	,'ALMProject.DefaultTask.Keywords'
	,'ALMProject.DefaultTask.Submitter'
	,'ALMProject.DefaultTask.Originator');


	eval {$entityobjs = $session->CopyEntity($projEntity, \@copyArray, \@skipArray);};
	if ($@) {
	 $session->OutputDebugString("յ CopyEntity  $@\n");
	 CQPerlExt::CQSession_Unbuild($session);
	 die "յ CopyEntity  $@";
	}

	$entsCount = $entityobjs->Count();
	$session->OutputDebugString(" CopyEntity дʵ = $entsCount\n");

	#################################################################################################################
	#When you get an error like this means the unique name is becoming duplicate.....
	#Call CommitEntities
	#Got CommitEntities Error Failed condition: GetState() == AD_VALIDATED
	#Location: ClearQuest Core:adeditaction.cpp:742
	#GetState()=PREPARED at h:/bugbert/nt_i386/pvtinstall/ClearQuest/lib/perl5/site_perl/5.8.6/CQPerlExt.pm line 43.
	##################################################################################################################
	## If you are copying a stateless record make sure the keys are unique and will not cause duplicates.
	for ($idx=0; $idx < $entsCount; $idx++) {

		$ent = $entityobjs->Item ($idx);
		my $entdefname = $ent->GetEntityDefName();
		my $dspName = $ent->GetDisplayName();
		$session->OutputDebugString ("ʵ DefName Ϊ$entdefname \n DisplayName Ϊ$dspName\n");

		if($entdefname eq "ALMProject")
		{	
			$projName=$ent->GetFieldStringValue("Name");
			$clonedProjID=$ent->GetFieldStringValue("id");
			($opts{clonedprojname} ne "") ?  ($clonedProjName=$opts{clonedprojname}) : ($clonedProjName="CopyOf_${projectID}_" . time());
			$failure=$ent->SetFieldValue("Name","$clonedProjName");
			if ($opts{category} ne ""){$failure=$ent->SetFieldValue("Category","$opts{category}");}
			if ($opts{release} ne ""){$failure=$ent->SetFieldValue("Release","$opts{release}");}
			$clonedprojSecPol=$ent->GetFieldStringValue("SecurityPolicy");
			$failure=$ent->SetFieldValue("Obsolete","No");
			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("StartDate","");
			$failure=$ent->SetFieldValue("EndDate","");
			$failure=$ent->SetFieldValue("DefaultRequest","");
			$failure=$ent->SetFieldValue("DefaultTask","");
			$failure=$ent->SetFieldValue("StatusLabel","");
			$failure=$ent->SetFieldValue("ReportTagsLabel","");
			$failure=$ent->SetFieldValue("CopyToProjectName","");
			$failure=$ent->SetFieldValue("CopyToCategory","");
			$failure=$ent->SetFieldValue("CopyToRelease","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
			$failure=$ent->SetFieldValue("OldID","$clonedProjID");
	
		}
		if($entdefname eq "ALMPhase")
		{	
		
			$priorPhaseLabel=$ent->GetFieldStringValue("PriorPhaseLabel");
			if ($priorPhaseLabel) {
			   	$APHash{$dspName}{'PriorPhaseLabel'}=$priorPhaseLabel;
			}
			if ($opts{category} ne ""){
				$failure=$ent->SetFieldValue("SecurityPolicy","$clonedprojSecPol");
			}
			$failure=$ent->SetFieldValue("PriorPhaseLabel","");
			$failure=$ent->SetFieldValue("PriorPhase","");
			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("StartDate","");
			$failure=$ent->SetFieldValue("EndDate","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");

		}
	
		if($entdefname eq "ALMRole")
		{
			$origdspName = $dspName;
			$origdspName =~ s/$clonedProjID/$projectID/;
			$AREntity = $session->GetEntity("ALMRole","$origdspName");
			$ARHash{$dspName}{'Members'} = 	$AREntity->GetFieldStringValue("Members");
			$ARHash{$dspName}{'Groups'} = 	$AREntity->GetFieldStringValue("Groups");
			$ARHash{$dspName}{'ApprovedActions'} = $AREntity->GetFieldStringValue("ApprovedActions");
			if ($opts{category} ne ""){
				$failure=$ent->SetFieldValue("SecurityPolicy","û");
			}
			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("EntityName","");
			$failure=$ent->SetFieldValue("Members","");
			$failure=$ent->SetFieldValue("ApprovedActions","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
		
		}
		if($entdefname eq "ALMWorkConfiguration")
		{

			$AWCHash{$dspName}{'ChildrenTypes_primary'} = $ent->GetFieldStringValue("ChildrenTypes_primary");
			$AWCHash{$dspName}{'ChildrenTypes_secondary'} = $ent->GetFieldStringValue("ChildrenTypes_secondary");
			$AWCHash{$dspName}{'ALMRoles'} = $ent->GetFieldStringValue("ALMRoles");
			$AWCHash{$dspName}{'PrimaryRole'} = $ent->GetFieldStringValue("PrimaryRole");
			if ($opts{category} ne ""){
				$failure=$ent->SetFieldValue("SecurityPolicy","$clonedprojSecPol");
			}
			$failure=$ent->SetFieldValue("ChildrenTypes_primary","");
			$failure=$ent->SetFieldValue("ChildrenTypes_secondary","");
			$failure=$ent->SetFieldValue("ALMRoles","");
			$failure=$ent->SetFieldValue("PrimaryRole","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
		
		}
		if($entdefname eq "ALMRequest")
		{
			$clonedRequestID=$ent->GetFieldStringValue("id");
			if ($opts{category} ne ""){
				$failure=$ent->SetFieldValue("Category","$opts{category}");
	 			$failure=$ent->SetFieldValue("Project","$clonedProjID");
			}
			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("CategoryStringPath","");
			$failure=$ent->SetFieldValue("DueDate","");
			$failure=$ent->SetFieldValue("StatusLabel","");
			$failure=$ent->SetFieldValue("Originator","");
			$failure=$ent->SetFieldValue("Keywords","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
			$failure=$ent->SetFieldValue("OldID","$clonedRequestID");
			$failure=$ent->SetFieldValue("Submitter","$currUser");
	 		$failure=$ent->SetFieldValue("SubmitDate", GetCurrentDate());
				
		}
		if($entdefname eq "ALMTask")
		{
			$clonedTaskID=$ent->GetFieldStringValue("id");
			$taskType=$ent->GetFieldStringValue("Type");
			if ($opts{category} ne ""){
				$failure=$ent->SetFieldValue("Project","$clonedProjID");
				$failure=$ent->SetFieldValue("SecurityPolicy","$clonedprojSecPol");
			}
			if ($clonedRequestID){
				$failure=$ent->SetFieldValue("Request","$clonedRequestID");
 			}
			else {
				CQPerlExt::CQSession_Unbuild($session);
				die "*** Ŀ $projectID  DefaultRequest";
			}
			$failure=$ent->SetFieldValue("Description","");
			$failure=$ent->SetFieldValue("ResolutionSummary","");
			$failure=$ent->SetFieldValue("Owner","");
			$failure=$ent->SetFieldValue("Type","");
			$failure=$ent->SetFieldValue("ResolutionCode","");
			$failure=$ent->SetFieldValue("DueDate","");
			$failure=$ent->SetFieldValue("StatusLabel","");
			$failure=$ent->SetFieldValue("Originator","");
			$failure=$ent->SetFieldValue("Keywords","");
			$failure=$ent->SetFieldValue("ratl_mastership","$localreplica");
			$failure=$ent->SetFieldValue("OldID","$clonedTaskID");
			$failure=$ent->SetFieldValue("Submitter","$currUser");
	 		$failure=$ent->SetFieldValue("SubmitDate", GetCurrentDate());
				
		}
	}

	$session->OutputDebugString(" ValidateEntities\n");

	#catch exceptions with eval
	############################

	eval{$strarray1 = $session->ValidateEntities($entityobjs);};
	my $size = grep(/\S/, @$strarray1); 
	if  ($@ || $size) {
			$session->RevertEntities($entityobjs);
			$ent=undef;
			$entityobjs=undef;
			CQPerlExt::CQSession_Unbuild($session);
			die "**  ValidateEntities յ$@ \nص״̬Ϊ@$strarray1\n";		
	}
	else
	{
		$session->OutputDebugString(" CommitEntities\n");
		eval {	$strarray2  = $session->CommitEntities($entityobjs);};
		my $size = grep(/\S/, @$strarray2);
		if  ($@ || $size)  
		{
			$session->RevertEntities($entityobjs);
			$ent=undef;
			$entityobjs=undef;
			CQPerlExt::CQSession_Unbuild($session);
			die "**  CommitEntities յ$@ \nص״̬Ϊ@$strarray2\n";
		}
		else{
			$session->OutputDebugString("CopyEntity ѳɹ\n");
			$ent=undef;
			$entityobjs=undef;
		}
	}

	######################
	# Restore the values for ALMPhase record on pass2 
	$session->OutputDebugString("ڵڶָ׶Ρ¼ֵ...\n") if (keys %APHash);
	$refh_AP=\%APHash;
	foreach my $AP ( keys %$refh_AP ) {
       
	       $priorPhaseLabel="";
	       $priorPhaseLabel =  $refh_AP->{$AP}->{'PriorPhaseLabel'};
	       $session->OutputDebugString("ALMPhase Ϊ$AP\n");
	       if ($priorPhaseLabel) {
		 		$session->OutputDebugString("PriorPhaseLabel Ϊ $priorPhaseLabel\n");
		 		$APEntity = $session->GetEntity("ALMPhase","$AP");
		 		$APEntity->EditEntity("modify");
		 		$failure=$APEntity->SetFieldValue("PriorPhaseLabel","$priorPhaseLabel");
	       }
	       CQValidateCommit($APEntity);
 
	}
	######################
	# Restore the values for WorkConfigurations fields on pass2 
	$session->OutputDebugString("ڵڶָ WorkConfigurations ֶεֵ...\n") if (keys %AWCHash);
	$refh_AWC=\%AWCHash;
	foreach my $AWC ( keys %$refh_AWC ) {
     
	       $primary =  $refh_AWC->{ $AWC }->{'ChildrenTypes_primary'};
	       $secondary =  $refh_AWC->{ $AWC }->{'ChildrenTypes_secondary'}; 
	       $roles =  $refh_AWC->{ $AWC }->{'ALMRoles'};
	       $primaryRole= $refh_AWC->{ $AWC }->{'PrimaryRole'};
	       $session->OutputDebugString("ALMWC Ϊ$AWC\n");
	       if ($primary || $secondary || $roles || $primaryRole) {
		 		$session->OutputDebugString("򡰽ɫֵ\n");
		 		$AWCEntity = $session->GetEntity("ALMWorkConfiguration","$AWC");
		 		$AWCEntity->EditEntity("modify");
		 		if ($primary) {
	       	 		@ptypes = split(/\n/, $primary);
	                		$session->OutputDebugString("ѸƵӴΪ@ptypes\n");
		 			foreach (@ptypes) {
						$failure=$AWCEntity->AddFieldValue("ChildrenTypes_primary","$_");
	 
	         			}  
		 		}
	         		if ($secondary) {
		       	   	@stypes = split(/\n/, $secondary);
			 		$session->OutputDebugString("ѸƵĸӴΪ@stypes\n");
			 		foreach (@stypes) {
						$failure=$AWCEntity->AddFieldValue("ChildrenTYpes_secondary","$_");
		 	 		} 
	         		}
		 	 	if ($roles) {
	       	   	 	@AWCroles = split(/\n/, $roles);
			 	 	$session->OutputDebugString("ɫΪ@AWCroles\n");
				 	foreach (@AWCroles) {
						s/$projectID/$clonedProjID/;
						$session->OutputDebugString("ѸƵĽɫΪ$_\n");
						$failure=$AWCEntity->AddFieldValue("ALMRoles","$_");
		 	 		} 
	         		}
	         		if ($primaryRole) {
		 	 	 	$session->OutputDebugString("PrimaryRole Ϊ$primaryRole\n");
      			 	$primaryRole =~ s/$projectID/$clonedProjID/;
		 		 	$session->OutputDebugString("ѸƵ PrimaryRole Ϊ$primaryRole\n");
		 		 	$failure=$AWCEntity->SetFieldValue("PrimaryRole","$primaryRole");
	        		}
		 	 	CQValidateCommit($AWCEntity);
 
	   	}

	}

	######################
	# Restore the values for ALMRoles fields on pass2 
	$session->OutputDebugString("ڵڶָɫֶεֵ...\n") if (keys %ARHash);
	$refh_AR=\%ARHash;
	foreach my $AR ( keys %$refh_AR ) {

	       $members =  $refh_AR->{$AR}->{'Members'};
	       $groups  =  $refh_AR->{$AR}->{'Groups'};	
	       $actions =  $refh_AR->{$AR}->{'ApprovedActions'};
	       $session->OutputDebugString("ALMRole Ϊ$AR\n");
	       if ($members || $actions || $groups) {
		 		$session->OutputDebugString("Members ֵ\n");
		 		$AREntity = $session->GetEntity("ALMRole","$AR");
		 		$AREntity->EditEntity("modify");
	         		if ($members) {
		 			@ARMembers = split(/\n/, $members);
					$session->OutputDebugString("Ա@ARMembers\n");
		 			foreach (@ARMembers) {
						$session->OutputDebugString("Ա$_\n");
						$failure=$AREntity->AddFieldValue("Members","$_");
		 			} 
	         		}
				if ($groups) {
		 			@ARGroups = split(/\n/, $groups);
					foreach (@ARGroups) {
						$session->OutputDebugString("$_\n");
						$failure=$AREntity->AddFieldValue("Groups","$_");
		 			} 
	        		}
		 		if ($actions) { 
	         		@ARActions = split(/ /, $actions);
		 			foreach (@ARActions) {
						$session->OutputDebugString("׼Ĳ\n$_\n");
						$failure=$AREntity->AddFieldValue("ApprovedActions","$_");
		 			}
	        		}
	        		CQValidateCommit($AREntity);
	       }
	}

	######################
	# Restore the values Type and Owner for DefaulTask
	if ($taskType) {
	 $session->OutputDebugString("ڵڶָ DefaultTask ֵ...\n");
	 $taskEntity = $session->GetEntity("ALMTask","$clonedTaskID");
	 $session->OutputDebugString("ALMTask Ϊ $taskType\n");
	 $taskEntity->EditEntity("modify");
	 $failure=$taskEntity->SetFieldValue("Type","$taskType");
	 CQValidateCommit($taskEntity);
	}

	######################
	#Restore the values for DefaultRequest and/or DefaultTask and SercurityPolicy fields on the project
	if ($clonedRequestID || $clonedTaskID) {
	 $session->OutputDebugString("ڵڶָĿ DefaultRequest  DefaultTask ֵ...\n");
	 $projEntity = $session->GetEntity("ALMProject","$clonedProjID");
	 $projEntity->EditEntity("modify");
	 if ($clonedRequestID){
	   $session->OutputDebugString("ڽ¡ĿȱʡΪ $clonedRequestID\n");
	   $failure=$projEntity->SetFieldValue("DefaultRequest","$clonedRequestID");
	 }
	 if ($clonedTaskID){
	   $session->OutputDebugString("ڽ¡ĿȱʡΪ $clonedTaskID\n");
	   $failure=$projEntity->SetFieldValue("DefaultTask","$clonedTaskID");
	 }
	 CQValidateCommit($projEntity);
	}
	$session->OutputDebugString("\nĿ $projectID ѳɹƵ $clonedProjName\n");
	$session->SetNameValue("CopyToProjectID", "$clonedProjID");
	$failure = $entity->SetFieldValue("CopyToProjectName","");
	$failure = $entity->SetFieldValue("CopyToCategory","");
	$failure = $entity->SetFieldValue("CopyToRelease","");

	sub CQValidateCommit {
	
		my $Entity=shift(@_);
		my $status;
		$session->OutputDebugString("֤ʵ\n");
		$status="";
		my $currEntityName = $Entity->GetDisplayName();
		my $currEntityType = $Entity->GetEntityDefName();
		eval {$status = $Entity->Validate()};

		if  ($@ || ($status ne "")) 
		{
			my $msg = "\n֤ʧ\n$@\nص״̬Ϊ$status\n$currEntityType$currEntityName\nѻԭ\n";
			$session->OutputDebugString("$msg");
		    	$Entity->Revert();
			CQPerlExt::CQSession_Unbuild($session);
			die "$msg";
		}
		else
		{
			$session->OutputDebugString("ʵʵ\n");
			eval {$status = $Entity->Commit()};
			if ($@ || ($status ne ""))
			{

				my $msg = "\nʵʧ\n$@\nص״̬Ϊ$status\n$currEntityType$currEntityName\nѻԭ\n";
				$session->OutputDebugString("$msg");
		    		$Entity->Revert();
				CQPerlExt::CQSession_Unbuild($session);
				die "$msg";
		
			}
			else {
				my $msg = "$currEntityType $currEntityName ʵ";
				$session->OutputDebugString("$msg");
			}
	
		}
		return $status;

	}
	sub GetCurrentDate 
	{
		my ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $time) = localtime();
		return sprintf("%4d-%2.2d-%2.2d %2.2d:%2.2d:%2.2d", $year + 1900, $mon + 1, $mday,  $hour, $min, $sec);
	}
	$session->SetNameValue("CopyToProjectID", "$clonedProjID");

	if (! $hasParam) # If called using the Project form's button
	{
		$failure = $entity->SetFieldValue("CopyToProjectName","");
		$failure = $entity->SetFieldValue("CopyToCategory","");
		$failure = $entity->SetFieldValue("CopyToRelease","");
		return "ͨƵǰĿ ALMProject $clonedProjID Ĵ";
	}
      return $result;
}
